/* Serial interface for AF_UNIX socket connections on Un*x like systems
   Copyright 1992, 1993, 1998 Free Software Foundation, Inc.

This file is part of GDB.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#include "defs.h"
#include "serial.h"
#include <sys/types.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
/* next line added by STAFF161 -- avoid bzero redef conflicts with socket.h */
#include <strings.h>
#include <sys/socket.h>

#ifndef __CYGWIN32__
/* next line added by STAFF161 -- force full sockaddr_un struct on DEC/OSF */
#define _XOPEN_SOURCE_EXTENDED 
#endif
#include <sys/un.h>

#include "signals.h"
#include "gdb_string.h"

struct unix_ttystate
{
  int bogus;
};

static int unix_open PARAMS ((serial_t scb, const char *name));
static void unix_raw PARAMS ((serial_t scb));
static int wait_for PARAMS ((serial_t scb, int timeout));
static int unix_readchar PARAMS ((serial_t scb, int timeout));
static int unix_setbaudrate PARAMS ((serial_t scb, int rate));
static int unix_setstopbits PARAMS ((serial_t scb, int num));
static int unix_write PARAMS ((serial_t scb, const char *str, int len));
/* FIXME: static void unix_restore PARAMS ((serial_t scb)); */
static void unix_close PARAMS ((serial_t scb));
static serial_ttystate unix_get_tty_state PARAMS ((serial_t scb));
static int unix_set_tty_state PARAMS ((serial_t scb, serial_ttystate state));
static int unix_return_0 PARAMS ((serial_t));
static int unix_noflush_set_tty_state PARAMS ((serial_t, serial_ttystate,
					      serial_ttystate));
static void unix_print_tty_state PARAMS ((serial_t, serial_ttystate));

void _initialize_ser_unix PARAMS ((void));

/* Open up a raw unix socket */

static int
unix_open(scb, name)
     serial_t scb;
     const char *name;
{
  char *port_str;
  int port;
  struct hostent *hostent;
  struct sockaddr_un sockaddr;
  int tmp;
  char hostname[100];
  struct protoent *protoent;
  int i;
  int len;

  port_str = strchr (name, ':');

  if (!port_str)
    error ("unix_open: No colon in host name!"); /* Shouldn't ever happen */

  tmp = min (port_str - name, (int) sizeof hostname - 1);
  strncpy (hostname, name, tmp); /* Don't want colon */
  hostname[tmp] = '\000';	/* Tie off host name */
  port_str++;                   /* advance port_str */

  for (i = 1; i <= 15; i++)
    {
      scb->fd = socket (PF_UNIX, SOCK_STREAM, 0);
      if (scb->fd < 0)
	return -1;
      sockaddr.sun_family = AF_UNIX;
      strcpy(sockaddr.sun_path, port_str);
      len = 
#if ! (defined(__linux__) || defined(__CYGWIN32__))
sizeof(sockaddr.sun_len) + 
#endif
sizeof(sockaddr.sun_family) +
                strlen(sockaddr.sun_path) + 1;
#if  ! (defined(__linux__) || defined(__CYGWIN32__))
      sockaddr.sun_len = len; 
#endif

      if (!connect (scb->fd, (struct sockaddr *) &sockaddr, len))
	break;

      close (scb->fd);
      scb->fd = -1;

/* We retry for ECONNREFUSED because that is often a temporary condition, which
   happens when the server is being restarted.  */

      if (errno != ECONNREFUSED)
	return -1;

      sleep (1);
    }

  tmp = 1;

  signal(SIGPIPE, SIG_IGN);	/* If we don't do this, then GDB simply exits
				   when the remote side dies.  */

  return 0;
}

static serial_ttystate
unix_get_tty_state(scb)
     serial_t scb;
{
  struct unix_ttystate *state;

  state = (struct unix_ttystate *)xmalloc(sizeof *state);

  return (serial_ttystate)state;
}

static int
unix_set_tty_state(scb, ttystate)
     serial_t scb;
     serial_ttystate ttystate;
{
  struct unix_ttystate *state;

  state = (struct unix_ttystate *)ttystate;

  return 0;
}

static int
unix_return_0 (scb)
     serial_t scb;
{
  return 0;
}

static void
unix_raw(scb)
     serial_t scb;
{
  return;			/* Always in raw mode */
}

/* Wait for input on scb, with timeout seconds.  Returns 0 on success,
   otherwise SERIAL_TIMEOUT or SERIAL_ERROR.

   For termio{s}, we actually just setup VTIME if necessary, and let the
   timeout occur in the read() in unix_read().
 */

static int
wait_for (scb, timeout)
     serial_t scb;
     int timeout;
{
  int numfds;
  struct timeval tv;
  fd_set readfds, exceptfds;

  FD_ZERO (&readfds);
  FD_ZERO (&exceptfds);

  tv.tv_sec = timeout;
  tv.tv_usec = 0;

  FD_SET(scb->fd, &readfds);
  FD_SET(scb->fd, &exceptfds);

  while (1)
    {
      if (timeout >= 0)
	numfds = select(scb->fd+1, &readfds, 0, &exceptfds, &tv);
      else
	numfds = select(scb->fd+1, &readfds, 0, &exceptfds, 0);

      if (numfds <= 0)
        {
	  if (numfds == 0)
	    return SERIAL_TIMEOUT;
	  else if (errno == EINTR)
	    continue;
	  else
	    return SERIAL_ERROR;	/* Got an error from select or poll */
        }

      return 0;
    }
}

/* Read a character with user-specified timeout.  TIMEOUT is number of seconds
   to wait, or -1 to wait forever.  Use timeout of 0 to effect a poll.  Returns
   char if successful.  Returns -2 if timeout expired, EOF if line dropped
   dead, or -3 for any other error (see errno in that case). */

static int
unix_readchar (scb, timeout)
     serial_t scb;
     int timeout;
{
  int status;

  if (scb->bufcnt-- > 0)
    return *scb->bufp++;

  status = wait_for(scb, timeout);

  if (status < 0)
    return status;

  while (1)
    {
      scb->bufcnt = read(scb->fd, scb->buf, BUFSIZ);
      if (scb->bufcnt != -1 || errno != EINTR)
	break;
    }

  if (scb->bufcnt <= 0)
    {
      if (scb->bufcnt == 0)
        return SERIAL_TIMEOUT;	/* 0 chars means timeout [may need to
				     distinguish between EOF & timeouts
				     someday] */
      else
        return SERIAL_ERROR;	/* Got an error from read */
    }

  scb->bufcnt--;
  scb->bufp = scb->buf;
  return *scb->bufp++;
}

static int
unix_noflush_set_tty_state (scb, new_ttystate, old_ttystate)
     serial_t scb;
     serial_ttystate new_ttystate;
     serial_ttystate old_ttystate;
{
  return 0;
}

static void
unix_print_tty_state (scb, ttystate)
     serial_t scb;
     serial_ttystate ttystate;
{
  /* Nothing to print.  */
  return;
}

static int
unix_setbaudrate(scb, rate)
     serial_t scb;
     int rate;
{
  return 0;			/* Never fails! */
}

static int
unix_setstopbits(scb, num)
     serial_t scb;
     int num;
{
  return 0;			/* Never fails! */
}

static int
unix_write(scb, str, len)
     serial_t scb;
     const char *str;
     int len;
{
  int cc;

  while (len > 0)
    {
      cc = write(scb->fd, str, len);

      if (cc < 0)
	return 1;
      len -= cc;
      str += cc;
    }
  return 0;
}

static void
unix_close(scb)
     serial_t scb;
{
  if (scb->fd < 0)
    return;

  close(scb->fd);
  scb->fd = -1;
}

static struct serial_ops unix_ops =
{
  "unix",
  0,
  unix_open,
  unix_close,
  unix_readchar,
  unix_write,
  unix_return_0, /* flush output */
  unix_return_0, /* flush input */
  unix_return_0, /* send break */
  unix_raw,
  unix_get_tty_state,
  unix_set_tty_state,
  unix_print_tty_state,
  unix_noflush_set_tty_state,
  unix_setbaudrate,
  unix_setstopbits,
  unix_return_0,	/* wait for output to drain */
};

void
_initialize_ser_unix ()
{
  serial_add_interface (&unix_ops);
}
